type formule = | Var of int
               | Et of formule * formule
               | Ou of formule * formule
               | Non of formule;;

(*1*)

let exemple_formule = Non (Et (Var 3, Ou (Var 1, Non (Var 2))));;



 (*2*)
 
 let rec evalue f contexte = match f with 
 | Var(i) -> contexte.(i) 
 | Et(p,q) -> (evalue p contexte) && (evalue q contexte)
 | Ou(p,q) -> (evalue p contexte) || (evalue q contexte)
 | Non(p) -> not(evalue p contexte);;
 
 (*3*)
let rec max_indice_var f = match f with 
  | Var i -> i
  | Non f1 -> max_indice_var f1
  | Et (f1, f2)  -> max (max_indice_var f1) (max_indice_var f2)
  | Ou (f1, f2) -> max (max_indice_var f1) (max_indice_var f2)
;;

(*4*)
let rec teste_depuis_indice k n contexte f =
  if k=n then evalue f contexte
  else
    begin
      contexte.(k+1) <- true;
      if teste_depuis_indice (k+1) n contexte f then true
      else
        begin
          contexte.(k+1) <- false;
          teste_depuis_indice (k+1) n contexte f
        end
    end
;;

(*5*)
let satisfaisable f =
  let n = max_indice_var f in
  let contexte = Array.make (n+1) true in
  let res = teste_depuis_indice 0 n contexte f in
    (res, contexte)
;;

satisfaisable exemple_formule;;




(*6*)
let implique p q = Ou(q, Non(p));;

let rec etliste l = match l with 
|[] -> failwith ""
|[x] -> x
| x :: q -> Et(x, etliste q);;

(* On considère les variables v1, v2, v3, v4 et v5 où 
v1 = "est ecossais"
v2 = "joue au golf"
v3 = "porte un kilt"
v4 = "sont mariés"
v5 = "sortent le dimanche"*)

let f1 = implique (Non(Var(1))) (Var(2));;

let f2 = Ou(Var(3),Non(Var(2)));;

let f3 = implique (Var(4)) (Non(Var(5)));;

let f4 = Et(implique (Var(4)) (Var(5)),implique (Var(5)) (Var(4)));;

let f5 = implique (Var(3)) (Et(Var(1),Var(4)));;
let f5bis = implique (Var(3)) (Ou(Var(1),Var(4)));;

let f6 = implique (Var(1)) (Var(3));;

let ftotal = etliste [f1; f2 ; f3 ; f4 ; f5 ; f6];;

satisfaisable ftotal;;

let ftotalbis = etliste [f1; f2 ; f3 ; f4 ; f5bis ; f6];;

satisfaisable ftotalbis;;

(*7*)

let rec descend_negations arbre = match arbre with
  | Et (f1, f2) -> Et (descend_negations f1, descend_negations f2)
  | Ou (f1, f2) -> Ou (descend_negations f1, descend_negations f2)
  | Non (Et (f1, f2)) ->
      Ou (descend_negations (Non f1), descend_negations (Non f2))
  | Non (Ou (f1, f2)) ->
      Et (descend_negations (Non f1), descend_negations (Non f2))
  | Non (Non f1) -> descend_negations f1
  | _ -> arbre
;;

(*8*)

let forme_normale_conjonctive f =
  let rec fnc f = match f with
    | Ou (f1, f2) -> 
        begin 
          let f1b = fnc f1 and f2b = fnc f2 in
            match (f1b, f2b) with
              | (Et (f3, f4), _) -> Et (fnc (Ou (f3, f2b)), fnc (Ou(f4, f2b)))
              | (_, Et (f3, f4)) -> Et (fnc (Ou (f1b, f3)), fnc (Ou (f1b, f4)))
              | _ -> Ou (f1b, f2b)
        end
    | Et (f1, f2) -> Et (fnc f1, fnc f2)
    | _ -> f
  in fnc (descend_negations f)
;;



let ex2 = Ou(Et(Et(Var(1),Var(2)),Var(3)),Et(Var(4),Var(5)));;
let ex3 = Ou(Et(Et(Var(1),Var(2)),Var(3)),Ou(Var(4),Var(5)));;

forme_normale_conjonctive ex2;;
forme_normale_conjonctive ex3;;

let tautologie_clause f =
  let n =  max_indice_var f in
  let vu = Array.make (n+1) 0 in
  let rec parcours f = match f with
    |Var(i) when vu.(i)= -1 -> true
    |Non(Var(i)) when vu.(i) =1 -> true
    |Var(i) when vu.(i) = 0 -> vu.(i) <- 1;false
    |Non(Var(i)) when vu.(i) = 0 -> vu.(i) <- -1;false
    |Ou(f1,f2) -> parcours f1 || parcours f2
    |_ -> false
  in parcours f;;


tautologie_clause ex3;;

let rec liste_clauses f = match f with
  |Et(f1,f2) -> (liste_clauses f1)@(liste_clauses f2)
  |_ -> [f];;

let ex4 = forme_normale_conjonctive exemple_formule;;

liste_clauses ex4;;

let tautologie f =
  let fnc = forme_normale_conjonctive f in
  let l = liste_clauses fnc in
  let rec taut_list l = match l with
    |[] -> true
    |c::q -> tautologie_clause c && taut_list q
  in taut_list l;;


let satisfaisable2 f =
  not(tautologie (Non(f)));;

satisfaisable2 ftotal;;